/*
* Simulate RIP (distance vector routing). Does not implement time outs
* and regular updates.
*/
#include "simul.h"
#define MAX_ROUTES 25
#define MAX_HOSTS  25

typedef struct {
  int routes;
  struct {
    int dest;      /* host */
    int nexthop;   /* next hop (RIP-2) */
    int metric;    /* distance */
  } route[MAX_ROUTES];
} rip_t;

/*
* Routing table maintained per host.
*/
typedef struct {
  int neighbors;            /* number of neighbors */
  int neighbor[MAX_HOSTS];  /* neighbors of this host */
  rip_t r;
} host_t;

static host_t host[MAX_HOSTS];

/*
* Cost of link link[from][to].
*/
static int link[MAX_HOSTS][MAX_HOSTS];


/*
* Initialize link metrics.
*/
static links(void)
{
  link[0][1] = 1;
  link[0][3] = 1;
  link[1][0] = 1;
  link[1][4] = 1;
  link[1][2] = 1;
  link[2][1] = 1;
  link[2][4] = 10;
  link[3][0] = 1;
  link[3][4] = 1;
  link[4][1] = 1;
  link[4][2] = 10;
  link[4][3] = 1;
} /* links */


static void rip_show(int myself, int dest, int seq, rip_t *r)
{
  host_t *h = &host[myself];
  int i, k;

  printf("%0.2f: %c sends #%d to %c: ",
    tnow(), 'A' + myself, seq, 'A' + dest);
  for (k = 0; k < r->routes; k++) {
    printf("(%c,%d) ", 'A' + r->route[k].dest, r->route[k].metric);
  }
  printf("\n");
}


/*
* Host routine; updates routing table on receipt of message.
* If entry changes, generates a 'triggered update'.
*/
static void rip_in(EntityType e)
{
  int i, k;
  rip_t *r = (rip_t *)edata(e);
  host_t *h;            /* current host */
  int source = A(e,0);  /* source of routing message */
  int myself = A(e,1);  /* ourselves */
  int dest;             /* destination of route */
  rip_t new;            /* routing update */

  h = &host[(int)A(e,1)];
  new.routes = 0;
  for (i = 0; i < r->routes; i++) {
    dest = r->route[i].dest;
    /* find entry in host routing table */
    for (k = 0; k < h->r.routes; k++) {
      if (h->r.route[k].dest == dest) break;
    }

    /* Update if new metric or better than existing one. */
    if (k >= h->r.routes ||
        (r->route[i].metric + link[myself][source] < h->r.route[k].metric)) {
      h->r.route[k].metric  = r->route[i].metric + link[myself][source];
      h->r.route[k].nexthop = source;
      h->r.route[k].dest    = dest;
      new.route[new.routes] = h->r.route[k];
      new.routes++;
      printf("%0.2f (#%d): %c -> %c: %c (%d)\n", tnow(), id(e),
		'A' + myself, 'A' + dest, 'A' + h->r.route[k].nexthop,
        h->r.route[k].metric);
      if (k >= h->r.routes) h->r.routes = k+1;
    } 
  }
  dispose(e);

  if (new.routes) {
    /* add local route (should be doing this only initially) */
    new.route[new.routes++] = h->r.route[0];

    /* for each outgoing link */
    for (i = 0; i < h->neighbors; i++) {
      e = create_();
      edata_set(e, new, sizeof(new.routes) + new.routes*sizeof(new.route[0]));
      seta(e, 0, myself);         /* source */
      seta(e, 1, h->neighbor[i]); /* destination */
      /* schedule arrival at neighbor one tick after current time */
      sched(e, rip_in, 1.0 + ((double)i)/100.);
      rip_show(myself, h->neighbor[i], id(e), &new);
    }
  }
}


/*
* Update: send routing table from 'src' to its neighbors.
*/
static void rip_update(int src)
{
  rip_t r;
  int i, k;
  host_t *h;
  EntityType e;

  h = &host[src];

  /* send to each outgoing link */
  for (i = 0; i < h->neighbors; i++) {
    e = create_();
    edata_set(e, h->r, sizeof(h->r.routes) + h->r.routes*sizeof(h->r.route[0]));
    seta(e, 0, src);            /* source */
    seta(e, 1, h->neighbor[i]); /* destination */
    /* schedule arrival at neighbor one tick after current time */
    sched(e, rip_in, 1.0 + ((double)i)/100.);
    rip_show(src, h->neighbor[i], id(e), &h->r);
  }
} /* rip_update */


/*
* Initialize and start first update by first node updating neighbors.
*/
void prime(void)
{
  int i, j;
  EntityType e;
  rip_t r;

  for (i = 0; i < MAX_HOSTS; i++) {
    for (j = 0; j < MAX_HOSTS; j++) {
      link[i][j] = 0;
    }
  }

  /* initialize link metrics */
  links();

  /* create neighbor entries and local route */
  for (i = 0; i < MAX_HOSTS; i++) {
    host[i].r.routes = 1;
    host[i].r.route[0].dest = i;
    host[i].r.route[0].metric = 0;
    host[i].r.route[0].nexthop = i;
    host[i].neighbors = 0;
    for (j = 0; j < MAX_HOSTS; j++) {
      if (link[i][j] > 0) {
        host[i].neighbor[host[i].neighbors] = j;
        host[i].neighbors++;
      }
    }
  }

  rip_update(0);  /* first node updates neighbors */
}

void wrapup()
{
}
